/************************************************************
* fingshuf1. This program cuantitatively compares microbial
* community fingerprints based on DGGE or SSCP methods
* Please, check input file format in README.
* This is an effort to explain microbial diversity in nature 
* based on the results obtained from molecular methods.
*
* Program name: fingshuf1.c
*
* Date: September 5, 2006, 17 October 2006.
*
* Author: Juan M. Gonzalez
*
* Address: IRNAS-CSIC
*	   Avda. Reina Mercedes 10
*	   41012 - Sevilla, Spain
* E-mail: jmgrau@irnase.csic.es
*
* Build: gcc -o fingshuf1 fingshuf1.c -lgsl -lgslcblas -lm
* 	 Adding "-static" might be needed depending on your GSL installation
*
* Requirements: GNU Scientific Library (GSL)
*		(htp://www.gnu.org/software/gsl/)
*
************************************************************/
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <malloc.h>
/*--------------------------------------------------------*/
#define MAX_LINE 150	     /* Maximum length of an input line of data */
#define MAX_FILENAME 50      /* Maximum number of characters in filename */
#define ERROR 0.05	     /* error allowed for the detection of bands as equal */
#define MINGROUP 10	     /* value for the group with the lowest representation */
/*--------------------------------------------------------*/

struct ind {
	float m, d;	/* data for each individual migration, density, presence in other sample */
	};
typedef struct ind ind;

struct sres {
	float acxy, acyx;
	long num_x, num_y;
	};
typedef struct sres sres;

/**********************************************************/
    /* global permanent variables for using as command-line options */
long num_shuffle;  	/* Number of shuffles; -S */
long abundance;		/* Total abundance of bacteria (equal for X and Y); -A */
int copyright;		/* Printing copyright; -c */
float error;		/* error allowed for the detection of bands as equal */
int ming;		/* value for the group with the lowest representation */
/*********************************************************/
main(int argc, char *argv[])
{

		/* variables for the command line arguments */
	char *program_name;	/* name of the program collected from
				the command line (for errors?) */
	char command_line[MAX_LINE];	/* captures the full command line including options */
	int coms;		/* counter for number of options in command line */
	int z;			/* counter for capturing the full command line */

                /* timing related variables to CPU time */
	int time_option;	/* Timing is an option. Default is no = 0 */
        time_t time0, time1;    /* initial and final time */
        double run_time;        /* running time. Difference in secs of time0 */
	int t;		/* time in seconds to run the simulation */
	float tt;	/* actual time during the simulation in seconds */

		/* GSL; random generators */
	const gsl_rng_type *T;	/* GSL random number generator */
	gsl_rng *r;		/* GSL random number generator, seed */

	long i, j, k;		 /* counters */
	ind *indx; 	 /* information for each band or OTU in X*/
        ind *indy; 	 /* information for each band or OTU in Y*/
	long n_otux, n_otuy;	 /* number of bands or OTUs in X and Y */
	float *comx; 	 /* community of bacteria */
	float *comy;	 /* community of bacteria Y */
	long n_x, n_y;		 /* number of individuals (bacteria) in X and Y */	
        float *coma;     	 /* community of bacteria A equivalent to X for the random shuffling */
        float *comb;     	 /* community of bacteria B equivalent to Y for the random shuffling */
        long n_a, n_b;  	 /* number of individuals (bacteria) in A and B */
	float *samx;	 /* bacteria retrieved during a sampling effort of X */
	float *samy;	 /* bacteria retrieved during a sampling effort of Y */
	long s_x, s_y;	    	 /* number of individuals in a sample of X and Y, respectively */
	long n_s;		 /* number of samplings */
	float *cx;	 	 /* CX values for each subsampling */
        float *cxy;  	 /* CXY values for each subsampling */
        float *cy;   	 /* CY values for each subsampling */
        float *cyx;  	 /* CYX values for each subsampling */
        float acxy;  		 /* ACXY values for each series of subsampling */
        float acyx;  		 /* ACYX values for each series of subsampling */
	sres *result;	 /* gets results of acxy and acyx for each shuffle */
	float prob_x, prob_y;	 /* Probability of significant differences between X and Y */
	float e_probx, e_proby;	 /* Error of the probability of X and Y */
	sres result0;		 /* results of acxy, acyx, and their rankings for the original communities */
	int loops;		 /* number of samples to be used for the estimates */

			/* file names and related stuff */
	char file_in1[MAX_FILENAME];   /* Input filename 1 with the data for sample X: density, migration, presence */
        char file_in2[MAX_FILENAME];   /* Input filename 2 with the data for sample Y: density, migration, presence */
	char file_out[MAX_FILENAME];   /* Output file with results and estimations */
        FILE *data_file_in1;
	FILE *data_file_in2;
        FILE *data_file_out;

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	/* Declaration of functions */

	void print_result_shuffling(FILE *data_file_out, sres *result, long num_shuffle, float prob_x, float prob_y, float e_probx, float e_proby, sres result0);
	void get_rank(sres *result0, sres *result, long num_shuffle, float *prob_x_ptr, float *prob_y_ptr, float *e_probx_ptr, float *e_proby_ptr);
	void shuffle_communities(long n_x, float *comx, long n_y, float *comy, long *n_a_ptr, float *coma, gsl_rng *r_ptr);
	void print_original_data(FILE *data_file_out, float *cx, float *cxy, float *cy, float *cyx, float acxy, float acyx, long n_s, char command_line[]);
	void estimates(long n_1, float *com1, long n_2, float *com2, float *c1, float *c12, float *c2, float *c21, long *ns_ptr, float *ac12_ptr, float *ac21_ptr, long abundance, float error, gsl_rng *r_ptr); 
	void estimates_shuffle(long n_1, float *com1, long n_2, float *com2, float *ac12_ptr, float *ac21_ptr, long abundance, float error, gsl_rng *r_ptr);
        void generate_community(ind *ind, long n_otu, long *n_ptr, long abundance, float *com, int ming);
	void preread(FILE *data_file_in, long *n_otu_ptr, long *abundance_ptr);
        void read_input(FILE *data_file_in, long *n_otu_ptr, ind *indin);
	void usage(char *program_name);  /* tells how to use the command line arguments */
 	void print_cr(void);
	void memory_error(void);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    /* initializing global permanent variables for using as command-line options */

	copyright = 1;           /* printing copyright information */
	num_shuffle = 10000;     /* number of shuffles or replicates */
	time_option = 0;	 /* selects not to estimate CPU time by default */
	error = ERROR;		 /* Default error value for detecting bands as equal */
	ming = MINGROUP;	 /* Default value for ming */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	/* captures command line */
	coms = argc;
	(void)strcpy(command_line, argv[0]);
	z = 1;
	while (z < coms) {
                (void)strncat(command_line, " ", MAX_LINE);
		(void)strncat(command_line, argv[z], MAX_LINE);
		z = z + 1;
	}
	program_name = argv[0];

	/* command line arguments stuff */
        if (argc == 1) {
                (void)usage(program_name);
                exit(8);
        }
        while ((argc>1) && (argv[1][0] == '-')) {

                switch (argv[1][1]) {

                          /*
                          *  -I<file_name> input file name 2 (X)
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'I'
                          *    argv[1][2] starts the name
                          */
                        case 'I':
                          (void)strcpy(file_in1, &argv[1][2]);
                          break;

                          /*
                          *  -i<file_name> input file name 2 (Y)
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'i'
                          *    argv[1][2] starts the name
                          */
                        case 'i':
                          (void)strcpy(file_in2, &argv[1][2]);
                          break;
                                                                                                                  
                          /*
                          *  -O<file_name> output file name
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'O'
                          *    argv[1][2] starts the name
                          */
                        case 'O':
                          (void)strcpy(file_out, &argv[1][2]);
                          break;

			  /*
			  *   -S<integer> set the number of shuffles or replicates
			  *     argv[1][0] is the dash
			  *     argv[1][1] is the 'S'
			  *     argv[1][2] start the number
			  */ 
			case 'S':
			  num_shuffle = atol(&argv[1][2]);
			  break; 

                         /*
                          *   -E<integer> set the error
                          *     argv[1][0] is the dash
                          *     argv[1][1] is the 'E'
                          *     argv[1][2] start the number
                          */
                        case 'E':
                          error = atof(&argv[1][2]);
                          break;

                        /*
                          *   -M<integer> set the error
                          *     argv[1][0] is the dash
                          *     argv[1][1] is the 'M'
                          *     argv[1][2] start the number
                          */
                        case 'M':
                          ming = atoi(&argv[1][2]);
                          break;

			  /*
			  *   -t sets timing the program on
			  *     argv[1][0] is the dash
			  *     argv[1][1] is the 't'
			  *     No use of name or numbers
			  */
			case 't':
			  time_option = 1;
			  break;

			  /*
			  *   -A sets the bacterial abundance 
			  *     argv[1][0] is the dash
			  *     argv[1][1] is the 'A'
			  *     argv[1][2] starts the number
			  */
/*			case 'A':
			  abundance = atol(&argv[1][2]);
			  break;
 */                                                                                                                 
                          /*
                          *  -c avoid printing the copyright information
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'c'
                          *    argv[1][2] no use of any filename or number
                          */
                        case 'c':
                          copyright = 0;
                          break;

                          /*
                          *   -h displays help
                          *     argv[1][0] is the dash
                          *     argv[1][1] is the 'h'
                          *     argv[1][2] not used
                          */
                        case 'h':
                          (void)usage(program_name);
                          break;

               /* insert other options for the command line here */
 
			default:
			  (void) fprintf(stderr, "Bad option %s\n", argv[1]);  
			  usage(program_name);
		}  /* end of switch */

		/* move the argument list up one, move the count down one */
		argc--;
		argv++;
	}	/* end of while for command-line options */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

        /* print copyright */
        if (copyright == 1) {
                (void)print_cr();
        }
 
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	if (time_option == 1)
        time0 = time(NULL);     /* initial CPU time */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	/* initialize GSL random number generator */
	gsl_rng_env_setup();
	T = gsl_rng_default;
	r = gsl_rng_alloc(T);

	/* initialize the input and output files */
        data_file_in1 = fopen(file_in1, "r");   /* first input file */
        if (file_in1 == NULL) {
                (void)fprintf(stderr, "Input file 1 could not be opened.\n \n");
                exit(8);
        }
        data_file_in2 = fopen(file_in2, "r");   /* Second input file */
        if (file_in1 == NULL) {
                (void)fprintf(stderr, "Input file 2 could not be opened.\n \n");
                exit(8);
        }
	data_file_out = fopen(file_out, "w");

        /* preread input data to get the number of data points */
        (void)preread(data_file_in1, &n_otux, &abundance);
	abundance = abundance * ming;
/* test */ (void)fprintf(stderr, "Abundance 1 = %ld \n", abundance);

        /* close input files */
        (void)fclose(data_file_in1);
        /* open again input files to actually read input data */
        data_file_in1 = fopen(file_in1, "r");   /* first input file */

        /* read input file 1 and get bacterial community X */
        /* First allocate memory for the data */
        indx = (ind *)malloc(n_otux * sizeof(ind));
        assert(indx != NULL);
        comx = (float *)malloc(abundance * sizeof(float));
        assert(comx != NULL);

        (void)read_input(data_file_in1, &n_otux, indx);
        (void)generate_community(indx, n_otux, &n_x, abundance, comx, ming);

        /* preread input data to get the number of data points */
        (void)preread(data_file_in2, &n_otuy, &abundance);
        abundance = abundance * ming;
/* test */ (void)fprintf(stderr, "Abundance 2 = %ld \n", abundance);

        /* close input files */
        (void)fclose(data_file_in2);
        /* open again input files to actually read input data */
        data_file_in2 = fopen(file_in2, "r");   /* Second input file */

        /* read input file 2 and get bacterial community Y */
	/* First allocate memory for the data */
        indy = (ind *)malloc(n_otuy * sizeof(ind));
        assert(indy != NULL);
	comy = (float *)malloc(abundance * sizeof(float));
        assert(comy != NULL);

        (void)read_input(data_file_in2, &n_otuy, indy);
        (void)generate_community(indy, n_otuy, &n_y, abundance, comy, ming);

	/* deallocating unused memory */
        free(indx);
        free(indy);

        /* Close the input file since we are done reading */
        (void)fclose(data_file_in1);
	(void)fclose(data_file_in2);

	if (n_y > n_x) {
		abundance = n_y;
	}
	else {
		abundance = n_x;
	}
	/* estimate Cx, Cxy, Cy, Cyx, ACxy and ACyx for original DGGEs */
	/* first allocate memory to the array ptrs */
	loops = (10 * ((log(abundance)) - 1))/1;
	cx = (float *)malloc(loops * sizeof(float));
	assert(cx != NULL);
        cxy = (float *)malloc(loops * sizeof(float));
	assert(cxy != NULL);
        cy = (float *)malloc(loops * sizeof(float));
	assert(cy != NULL);
        cyx = (float *)malloc(loops * sizeof(float));
	assert(cyx != NULL);

	(void)estimates(n_x, comx, n_y, comy, cx, cxy, cy, cyx, &n_s, &acxy, &acyx, abundance, error, r);
        result0.acxy = acxy;   /* data from original community at the position 0 */
        result0.acyx = acyx;
	result0.num_x = 0;
	result0.num_y = 0;

	/* print results from original sample */
	(void)print_original_data(data_file_out, cx, cxy, cy, cyx, acxy, acyx, n_s, command_line);
	/* deallocate memory occuped by c's */
	free(cx);
	free(cxy);
	free(cy);
	free(cyx);

	/* allocate memory for ptrs */
	coma = (float *)malloc(n_x * sizeof(float));
	assert(coma != NULL);
	comb = (float *)malloc(n_y * sizeof(float));
	assert(comb != NULL);
	result = (sres *)malloc(num_shuffle * sizeof(sres));
	assert(result != NULL);
	i = 1;
	/* start shuffling the communities to build random communities */
	while (i < (num_shuffle+1)) {

/* test */ (void)fprintf(stderr, " Processing shuffle no. %ld \n", i);

		(void)shuffle_communities(n_x, comx, n_y, comy, &n_a, coma, r);
                (void)shuffle_communities(n_y, comy, n_x, comx, &n_b, comb, r);
		(void)estimates_shuffle(n_a, coma, n_b, comb, &acxy, &acyx, abundance, error, r);	

		result[i-1].acxy = acxy;
		result[i-1].acyx = acyx;
		result[i-1].num_x = i;
		result[i-1].num_y = i;
		i = i + 1;
	}
	/* deallocate not needed memory */
	free(coma);
	free(comb);
        free(comx);
        free(comy);

/* test */ malloc_stats();

	(void)get_rank(&result0, result, num_shuffle, &prob_x, &prob_y, &e_probx, &e_proby);
	(void)print_result_shuffling(data_file_out, result, num_shuffle, prob_x, prob_y, e_probx, e_proby, result0);

	/* deallocating */
	free(result);
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
                /* for CPU time */
        if (time_option == 1) {
              time1 = time(NULL);
              run_time = difftime(time1, time0);
	      if ((run_time/7200.0) < 1.0) {
		(void)fprintf(data_file_out, "\n Running time = %.2f minutes \n \n", (run_time/60.0));
	      }
	      else {
		(void)fprintf(data_file_out, "\n Running time = %.2f hours \n \n", (run_time/3600.0));
	      }
        }
/*_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _*/

        /* close output file */
        (void)fclose(data_file_out);

/* test */ malloc_stats();

	return;
	
}	/* End of main */

/****************************************************************
Preread input data to count number of data points.
*****************************************************************/
void preread(FILE *data_file_in,
                long *n_otu_ptr,
		long *abundance_ptr)
{

        long l; /* counter */
        float otu_m;   /* migration for each OTU */
        float ab;      /* density, abundance or proportion for each OTU */
        char line[MAX_LINE]; /* gets each line from input file */
        char *data_ptr;         /* character returned by fgets. NULL= end of file */
        float min;              /* minimum value of density */
        float sum;              /* sum of d */

        /* Data should be in the following format:
                # Comments
                # OTU Individuals
                3.1 24
                4.5 10
        Each line should have only these data separated by a single space.
        DO NOT leave any empty line at the end of the file */

        min = 100000.0;
        sum = 0.0;
        l = 0;          /* initialize number of OTUs */
        while (1) {
                data_ptr = fgets(line, sizeof(line), data_file_in);
                        /* check if we read every line */
                if (data_ptr == NULL)
                        break;
                        /* process new line */
                else if ((line[0] != '#') && (line[0] != '!')) {
                        /* read data line */
                        (void)sscanf(line, "%f %f", &otu_m, &ab);
                        l = l + 1;  /* a new OTU */
                        sum = sum + ab;
                        if (ab <= min) {
                                min = ab;
                        }
                }
        }
        *n_otu_ptr = l;

                /* round abundance up */
        *abundance_ptr = (long) (sum * (1.0/min));

        return;

}

/****************************************************************
Read input data: OTU and number of individuals.
OTU is given by its migration on a DGGE gel (in arbitrary units)
*****************************************************************/
void read_input(FILE *data_file_in,
		long *n_otu_ptr,
		ind *indin)
{

	long l;	/* counter */
	float otu_m;   /* migration for each OTU */
	float ab;      /* density, abundance or proportion for each OTU */
	char line[MAX_LINE]; /* gets each line from input file */
        char *data_ptr;         /* character returned by fgets. NULL= end of file */
	
        /* Data should be in the following format:
		# Comments
		# OTU Individuals
		3.1 24
		4.5 10
        Each line should have only these data separated by a single space.
	DO NOT leave any empty line at the end of the file */

	/* memory has been already allociated */
	
	l = 0; 		/* initialize number of OTUs */
        while (1) {
                data_ptr = fgets(line, sizeof(line), data_file_in);
                	/* check if we read every line */
                if (data_ptr == NULL)
                        break;
                  	/* process new line */
                else if ((line[0] != '#') && (line[0] != '!')) {
			/* read data line */
                        (void)sscanf(line, "%f %f", &otu_m, &ab);
			/* transfer info to pointer list with the data */
                	indin[l].m = otu_m;
                	indin[l].d = ab;
                        l = l + 1;  /* a new OTU */
		}
	}
	*n_otu_ptr = l;

	return;

}
/****************************************************************
Generate a microbial community from the input data of number of OTUs
and their number of individuals.
*****************************************************************/
void generate_community(ind *indin,
			   long n_otu,
			   long *n_ind_ptr,
			   long abundance,
			   float *com_ptr,
			   int ming)
{

	float sum_ind;		/* sum of abundances, densities or proportions */
	long p;			/* counter of individuals in the community */
	long q;			/* holds the number of otus */
	long t;			/* gets as an 'int' the number of individuals for an OTU */
	long s;			/* counter for com_ptr */
        float min;              /* minimum value of densities */
        float factor;           /* conversion factor for generating the community */

       /* get minimum value of density */
        min = 100000.0;
        q = n_otu;
        for (p = 0; p < q; p++) {
                if (indin[p].d <= min) {
                        min = indin[p].d;
                }
        }

        /* estimate abundance for each otu */
        factor = (ming*1.0)/min;
        s = 0;
        for (p = 0; p < q; p++) {
                indin[p].d = indin[p].d * factor;
        }

        /* estimate sum of d for number of otus */
        sum_ind = 0.0;
        for (p = 0; p < q; p++) {
                sum_ind = sum_ind + indin[p].d;
        }

        /* generate community from each OTU abundance */
        s = 0;
        for (p = 0; p < q; p++) {
                t = (long) (indin[p].d/1);
                while (t > 0) {
                        com_ptr[s] = indin[p].m;
                        s = s + 1;
                        t = t - 1;
                }
        }
        *n_ind_ptr = s;

/* test */ (void)fprintf(stderr, "Abundance = %ld; Counted individuals = %ld \n", abundance, s);

	return;
}

/****************************************************************
Estimates de parameters for the Cramer-von Mises statistic
*****************************************************************/
void estimates(	long	n_1,
		float	*com1,
		long	n_2,
                float   *com2,
		float	*c1_ptr,
		float	*c12_ptr,
		float	*c2_ptr,
		float	*c21_ptr,
		long	*ns_ptr,
		float	*ac12_ptr,
		float	*ac21_ptr,
		long	abundance,
		float   error,
		gsl_rng *r_ptr)
{

     float *sam1;   	 /* bacteria retrieved during a sampling effort of X */
     float *sam2;    	 /* bacteria retrieved during a sampling effort of Y */
/*     float *sam1b; */       /* a second sam1 to avoid cx = 1 */
/*     float *sam2b; */      /* a second sam2 to avoid cy = 1 */
     float t_ac12, t_ac21;  	 /* temporal ac12 and ac21 */
     long l;	 		 /* counter */
     float tc1, tc12, tc2, tc21; /* temporal variables for bringing c1, c12, c2, c21 */	
     int e;			 /* index for getting the value of l */
     int b;			 /* counter for the number of loops (index for c's) */
     int equal_samples;          /* signals if the samples to be compared are equal or not */

     void generate_sample(float *sam, long l, float *com1, long n_1, gsl_rng *r_ptr);
     void calculate_singletons(float *sam1, long s_a, float *sam2, long s_b, float *tc_ptr, float error, float equal_samples);

	/* allocate memory */
     sam1 = (float *)malloc(n_1 * sizeof(float));
     assert(sam1 != NULL);
     sam2 = (float *)malloc(n_2 * sizeof(float));
     assert(sam2 != NULL);
/*     sam1b = (float *)malloc(n_1 * sizeof(float));
     assert(sam1b != NULL);
     sam2b = (float *)malloc(n_2 * sizeof(float));
     assert(sam2b != NULL);
*/

	/* initializa counters and variables */
     t_ac12 = 0.0;
     t_ac21 = 0.0;
     b = 0;
     l = 0;
     e = 1;
     while (e <= ((log10(abundance))/1)) {   /* loops to get value of l */
	l = (pow(10, e))/1;
	while ((l < ((pow(10, e+1))/1)) && (l < n_1)) {
/*		l = l + (pow(10, e)/1);   */

     if ((l <= n_1) && (l <= n_2)) {
        (void)generate_sample(sam1, l, com1, n_1, r_ptr);
        (void)generate_sample(sam2, l, com2, n_2, r_ptr);

        equal_samples = 1;
        (void)calculate_singletons(sam1, l, sam1, l, &tc1, error, equal_samples);

        equal_samples = 0;
        (void)calculate_singletons(sam1, l, sam2, l, &tc12, error, equal_samples);

        equal_samples = 1;
        (void)calculate_singletons(sam2, l, sam2, l, &tc2, error, equal_samples);

        equal_samples = 0;
        (void)calculate_singletons(sam2, l, sam1, l, &tc21, error, equal_samples);

	/* transfer tc_s to pointer lists */
	c1_ptr[b] = tc1;
        c12_ptr[b] = tc12;
        c2_ptr[b] = tc2;
        c21_ptr[b] = tc21;
	/* calculate ac12 and ac21 */ 
	t_ac12 = t_ac12 + ((tc1 - tc12) * (tc1 - tc12));
        t_ac21 = t_ac21 + ((tc2 - tc21) * (tc2 - tc21));

	b = b + 1;
     } /* end of if l < n_1 and n_2 */ 

		l = l + (pow(10, e)/1);
     }  /* end loops to get the value of l */
     e = e + 1;
     }

	*ac12_ptr = t_ac12;
	*ac21_ptr = t_ac21;
	*ns_ptr = b;

        free(sam1);
        free(sam2);
/*	free(sam1b);
	free(sam2b);
*/

	return;

}

/****************************************************************
Estimates de parameters for the Cramer-von Mises statistic for the
shuffling approach
*****************************************************************/
void estimates_shuffle( long    n_1,
                	float  	*com1,
                	long    n_2,
                	float   *com2,
                	float   *ac12_ptr,
                	float   *ac21_ptr,
                	long    abundance,
                	float   error,
                	gsl_rng *r_ptr)
{

     float *sam1;            /* bacteria retrieved during a sampling effort of X */
     float *sam2;            /* bacteria retrieved during a sampling effort of Y */
     float t_ac12, t_ac21;       /* temporal ac12 and ac21 */
     long l;                     /* counter */
     float tc1, tc12, tc2, tc21; /* temporal variables for bringing c1, c12, c2, c21 */
     int e;                      /* index for getting the value of l */
/*     float *sam1b; */		/* second sam1 sample */
/*     float *sam2b; */		/* second sam2 sample */
    int equal_samples;         /* signals if the samples to be compared are equal or not */

     void generate_sample(float *sam_ptr, long l, float *com1_ptr, long n_1, gsl_rng *r_ptr);
     void calculate_singletons(float *sam1_ptr, long s_a, float *sam2_ptr, long s_b, float *tc1_ptr, float error, float equal_samples);
     void memory_error1(void);

        /* allocate memory */
     sam1 = (float *)malloc(n_1 * sizeof(float));
     assert(sam1 != NULL);
     sam2 = (float *)malloc(n_2 * sizeof(float));
     assert(sam2 != NULL);
/*     sam1b = (float *)malloc(n_1 * sizeof(float));
     assert(sam1b != NULL);
     sam2b = (float *)malloc(n_2 * sizeof(float));
     assert(sam2b != NULL);
*/

     /* initialize counters and variables */
     t_ac12 = 0.0;
     t_ac21 = 0.0;
     l = 0;
     e = 1;
     while (e <= ((log10(abundance))/1)) {   /* loops to get value of l */
	l = (pow(10, e)/1);
        while ((l < ((pow(10, e+1))/1)) && (l <= n_1)) {
/*                l = l + (pow(10, e)/1);   */

     if ((l <= n_1) && (l <= n_2)) {
        (void)generate_sample(sam1, l, com1, n_1, r_ptr);
        (void)generate_sample(sam2, l, com2, n_2, r_ptr);


        equal_samples = 1;
        (void)calculate_singletons(sam1, l, sam1, l, &tc1, error, equal_samples);

        equal_samples = 0;
        (void)calculate_singletons(sam1, l, sam2, l, &tc12, error, equal_samples);

        equal_samples = 1;
        (void)calculate_singletons(sam2, l, sam2, l, &tc2, error, equal_samples);

        equal_samples = 0;
        (void)calculate_singletons(sam2, l, sam1, l, &tc21, error, equal_samples);

        /* calculate ac12 and ac21 */
        t_ac12 = t_ac12 + ((tc1 - tc12) * (tc1 - tc12));
        t_ac21 = t_ac21 + ((tc2 - tc21) * (tc2 - tc21));

     } /* end of if l < n_1 and n_2 */
		l = l + (pow(10, e)/1);
     }  /* end loops to get the value of l */
     e = e + 1;
     }

        *ac12_ptr = t_ac12;
        *ac21_ptr = t_ac21;

	free(sam1);
	free(sam2);
/*        free(sam1b);
        free(sam2b);
*/

        return;

}

/****************************************************************
Generate samples from the communities  
*****************************************************************/
void generate_sample(float *sam_ptr,
		     long l,
		     float *com_ptr,
		     long n_gs,
		     gsl_rng *r_ptr)
{

	if (l <= n_gs) {
		(void)gsl_ran_choose(r_ptr, sam_ptr, l, com_ptr, n_gs, sizeof(float));
	}

	return;

}  

/****************************************************************
Calculate the number of singletons (only present once)
between a couple of communities
*****************************************************************/
void calculate_singletons(float *sama_ptr,
			 long s_a,
			 float *samb_ptr,
			 long s_b,
			 float *c_ptr,
			 float error,
			 int equal_samples)
{

	long singletons;	/* number of singletons */
	long p, q;		/* counters for loops */
	long repeat;		/* counts number of repeated individuals */


        if (equal_samples = 1) {  /* comparing a sample with itself */
        singletons = 0;
        for (p = 0; p < s_a; p++) {
                repeat = 0;
                for (q = 0; q < s_b; q++) {
                        if (((sama_ptr[p] + error) >= samb_ptr[q]) && ((sama_ptr[p] - error) <= samb_ptr[q])) {
                                repeat = repeat + 1;
                        }
                }
                if (repeat <= 1) {
                        singletons = singletons + 1;
                }
        }
        }
        else { /* different samples */
        singletons = 0;
        for (p = 0; p < s_a; p++) {
                repeat = 0;
                for (q = 0; q < s_b; q++) {
                        if (((sama_ptr[p] + error) >= samb_ptr[q]) && ((sama_ptr[p] - error) <= samb_ptr[q])) {
                                repeat = repeat + 1;
                        }
                }
                if (repeat == 0) {
                        singletons = singletons + 1;
                }
        }
        }

	*c_ptr = 1 - ((singletons * 1.0)/(s_a * 1.0));

	return;

}
/****************************************************************
Print the data obtained from the original (input) communities/samples
*****************************************************************/
void print_original_data(FILE *data_file_out,
			 float *cx_ptr,
			 float *cxy_ptr,
			 float *cy_ptr,
			 float *cyx_ptr,
			 float acxy,
			 float acyx,
			 long n_s,
			 char command_line[MAX_LINE])
{

	long w;			/* counter */

	(void)fprintf(data_file_out, "# Results of running: %s \n \n", command_line);
        (void)fprintf(data_file_out, "# Results from the original detected communities. \n \n");
	(void)fprintf(data_file_out, "# SAMPLE  CX     CXY     CY     CYX     (CX-CXY)2   (CY-CYX)2 \n");
        (void)fprintf(data_file_out, "# ------------------------------------------------------------------- \n");
	for (w = 0; w < n_s; w++) {
		(void)fprintf(data_file_out, "%ld  %.5f %.5f %.5f %.5f ", (w+1), cx_ptr[w], cxy_ptr[w], cy_ptr[w], cyx_ptr[w]);
		(void)fprintf(data_file_out, "%.5f %.5f \n", ((cx_ptr[w] - cxy_ptr[w])*(cx_ptr[w] - cxy_ptr[w])), ((cy_ptr[w] - cyx_ptr[w])*(cy_ptr[w] - cyx_ptr[w])));
	}
	(void)fprintf(data_file_out, "---------------------------------------------------------------------\n");
	(void)fprintf(data_file_out, "Acxy = %.5f;      	Acyx = %.5f \n", acxy, acyx);
	(void)fprintf(data_file_out, "=====================================================================\n \n");

	return;

}

/****************************************************************
This function shuffles the communities and picks a subsample of the
indicated number of individuals.
*****************************************************************/
void shuffle_communities(long n_t1,
			  float *com1_ptr,
			  long n_t2,
			  float *com2_ptr, 
			  long *n_s_ptr,
			  float *coms_ptr,	
			  gsl_rng *r_ptr)
{

	long k;				/* counter for loops */
	long t;				/* counter for number of elements in com1_ptr */
	float *temp;			/* holds community */
	long tt;			/* counter for total number of both communities */
	
        /* transfering community to temp array */
        temp = (float *)malloc((n_t1 + n_t2) * sizeof(float));
        assert(temp != NULL);

        t = 0; /* counts the number of elements */
        for (k = 0; k < n_t1; k++) {
                temp[k] = com1_ptr[k];
                t = t + 1;
        }
	tt = t;
	*n_s_ptr = t;
        for (k = 0; k < n_t2; k++) {
                temp[t+k] = com2_ptr[k];
                tt = tt + 1;
        }

	/* shuffle the total community form by both communities X + Y */
	(void)gsl_ran_shuffle(r_ptr, temp, tt, sizeof(float));

	/* transfer shuffled first t elements from temp list to output list */
	for (k = 0; k < t; k++) {
		coms_ptr[k] = temp[k];
	}

        free(temp);

	return;

}

/****************************************************************
Printing the ACXY and ACYX for each shuffle and the probability of significance
*****************************************************************/
void print_result_shuffling(FILE *data_file_out,
			    sres *result_ptr,
			    long num_shuffle,
			    float prob_x,
			    float prob_y,
			    float e_probx,
			    float e_proby,
			    sres result0)
{

	long w;		/* counter for ranks */
	long v;		/* counter for going through the shuffles */
	int data_0_done;	/* signals if original data have been printed out */

	(void)fprintf(data_file_out, "# Results of %ld shuffles of the given communities. \n", num_shuffle);
        (void)fprintf(data_file_out, "# Shuffling based on X \n");
	(void)fprintf(data_file_out, "# Rank(x)    ACxy     ACyx     Rank(y)\n");
        (void)fprintf(data_file_out, "# -------------------------------------------------------------------\n");

	data_0_done = 0;
	for (w = 0; w < num_shuffle; w++) {
                if (result0.num_x == w) {
                        (void)fprintf(data_file_out, "%ld %.5f %.5f %ld (original data) \n", (result0.num_x + 1), result0.acxy, result0.acyx, (result0.num_y + 1));
			data_0_done = 1;
                }
		for (v = 0; v < num_shuffle; v++) {	
                        if (result_ptr[v].num_x == w) {
                                (void)fprintf(data_file_out, "%ld %.5f %.5f %ld \n", (result_ptr[v].num_x + 1), result_ptr[v].acxy, result_ptr[v].acyx, (result_ptr[v].num_y + 1));
                        }
		}
	}
	if (data_0_done == 0) {
		(void)fprintf(data_file_out, "%ld %.5f %.5f %ld (original data) \n", (result0.num_x + 1), result0.acxy, result0.acyx, (result0.num_y + 1));
	}
        (void)fprintf(data_file_out, "#--------------------------------------------------------------------\n");
        (void)fprintf(data_file_out, "#   Probability XY = %.8f (error = %.8f ) \n", prob_x, e_probx);
        (void)fprintf(data_file_out, "#====================================================================\n");

	/* Now, process the Y based shuffle ranking */
        (void)fprintf(data_file_out, "# Shuffling based on Y \n");
        (void)fprintf(data_file_out, "# Rank(y)    ACyx     ACxy     Rank(x)\n");
        (void)fprintf(data_file_out, "# -------------------------------------------------------------------\n");

	data_0_done = 0;
        for (w = 0; w < num_shuffle; w++) {
                if (result0.num_y == w) {
                        (void)fprintf(data_file_out, "%ld %.5f %.5f %ld (original data) \n", (result0.num_y + 1), result0.acyx, result0.acxy, (result0.num_x + 1));
			data_0_done = 1;
                }
                for (v = 0; v < num_shuffle; v++) {
                        if (result_ptr[v].num_y == w) {
                                (void)fprintf(data_file_out, "%ld %.5f %.5f %ld \n", (result_ptr[v].num_y + 1), result_ptr[v].acyx, result_ptr[v].acxy, (result_ptr[v].num_x + 1));
                        }
                }
        }
	if (data_0_done == 0) {
		(void)fprintf(data_file_out, "%ld %.5f %.5f %ld (original data) \n", (result0.num_y + 1), result0.acyx, result0.acxy, (result0.num_x + 1));
	}
        (void)fprintf(data_file_out, "#--------------------------------------------------------------------\n");
        (void)fprintf(data_file_out, "#   Probability YX = %.8f (error = %.8f) \n", prob_y, e_proby);
        (void)fprintf(data_file_out, "#====================================================================\n");

	return;

}

/****************************************************************
Obtaining the rank where the original communites get placed among
the results from shuffling
*****************************************************************/
void get_rank(	    sres *result0_ptr,
	      	    sres *result_ptr,
	      	    long num_shuffle,
	      	    float *prob_x_ptr,
	      	    float *prob_y_ptr,
	      	    float *e_probx_ptr,
	      	    float *e_proby_ptr)
{

	long p, q, w;			/* counters */
	long rank;			/* rank or order in a list */
	sres *temp;			/* temporal list */
	sres *ordered;			/* ordered list */
  	long size_ord;			/* indicates the number of elements in ordered list */
	float prob;			/* temporal transference of probability values */
	long i_min, i_max;		/* indexes for min and max of a list */
	float min, max;			/* min and max in the list */
	int found;			/* indicates if a rank has been found */

	/* copy list */
	temp = (sres *)malloc(num_shuffle * sizeof(sres));
	assert(temp != NULL);
	for(p = 0; p < num_shuffle; p++) {
		temp[p].acxy = result_ptr[p].acxy;
		temp[p].num_x = result_ptr[p].num_x;
		temp[p].acyx = result_ptr[p].acyx;
		temp[p].num_y = result_ptr[p].num_y;
	}

	/* allocate memory */	
	ordered = (sres *)malloc(num_shuffle * sizeof(sres));
	assert(ordered != NULL);

	/* get max and min for acxy */
	i_min = 0;
	i_max = num_shuffle - 1;
	min = 100.0;
	max = -1.0;
	for(p = 0; p < num_shuffle; p++) {
		if (temp[p].acxy < min) {	
			min = temp[p].acxy;
			i_min = p;	
		}
		if (temp[p].acxy > max) {
			max = temp[p].acxy;
			i_max = p;
		}
	}

	/* place min and max in its place and rearrange temp */
	ordered[0].acxy = temp[i_min].acxy;
        ordered[0].num_x = temp[i_min].num_x;
        ordered[0].acyx = temp[i_min].acyx;
        ordered[0].num_y = temp[i_min].num_y;
        ordered[1].acxy = temp[i_max].acxy;
        ordered[1].num_x = temp[i_max].num_x;
        ordered[1].acyx = temp[i_max].acyx;
        ordered[1].num_y = temp[i_max].num_y;
	size_ord = 2;
	for (p = 0; p < (num_shuffle-1); p++) {
		if (p >= i_min) {
			temp[p].acxy = temp[p+1].acxy;
                        temp[p].num_x = temp[p+1].num_x;
                        temp[p].acyx = temp[p+1].acyx;
                        temp[p].num_y = temp[p+1].num_y;
		}
	}
	for (p = 0; p < (num_shuffle-2); p++) {
		if (p >= i_max) {
                        temp[p].acxy = temp[p+1].acxy;
                        temp[p].num_x = temp[p+1].num_x;
                        temp[p].acyx = temp[p+1].acyx;
                        temp[p].num_y = temp[p+1].num_y;
		}
	}
	/* place the rest of elements in their order */
	for (p = 0; p < (num_shuffle-2); p++) {
		q = 0;
		while (q <= (size_ord-2)) {
			if ((ordered[q].acxy < temp[p].acxy) && (ordered[q+1].acxy >= temp[p].acxy)) {
				/* insert element since we found its place */
				size_ord = size_ord + 1; /* list increases by 1 element */
				for (w = (size_ord-1); w > (q+1); w--) { /* rearrange ordered list */ 
					ordered[w].acxy = ordered[w-1].acxy;
				        ordered[w].num_x = ordered[w-1].num_x;
        				ordered[w].acyx = ordered[w-1].acyx;
        				ordered[w].num_y = ordered[w-1].num_y;
				}
				/* insert element in its position */	
			        ordered[q+1].acxy = temp[p].acxy;
			        ordered[q+1].num_x = temp[p].num_x;
			        ordered[q+1].acyx = temp[p].acyx;
			        ordered[q+1].num_y = temp[p].num_y;

				break;
			}
			q = q + 1;
		}
	}
	/* place order in num_x */
	for (p = 0; p < num_shuffle; p++) {
		ordered[p].num_x = (num_shuffle - 1) - p;
	}

	/* get the rank of the original rank */
	rank = 0;
	result0_ptr->num_x = rank; /* initialize at the begining of the list */
	found = 0;
	for (p = 0; p < num_shuffle; p++) {
		if ((result0_ptr->acxy >= ordered[p].acxy) && (found == 0)) {
			rank = (num_shuffle - 1) - p;
			result0_ptr->num_x = rank;
		}
		else {
			found = 1;
		}
	}

	/* transfer ordered list to temp */
	for (p = 0; p < num_shuffle; p++) {
		temp[p].acxy = ordered[p].acxy;
		temp[p].num_x = ordered[p].num_x;
                temp[p].acyx = ordered[p].acyx;
                temp[p].num_y = ordered[p].num_y;
	}

        /* calculate the probability and its error based on its rank */
	if (rank == 0) rank = 1;
        *prob_x_ptr = ((rank * 1.0)/((num_shuffle + 1) * 1.0));
	prob = *prob_x_ptr;
        *e_probx_ptr = ((prob * (1.0 - prob))/(num_shuffle * 1.0));

        /* start ordering in base to ACyx */
        /* get max and min for acxy */
        max = -1.0;
        min = 100.0;
	i_min = 0;
	i_max = num_shuffle - 1;
        for(p = 0; p < num_shuffle; p++) {
                if (temp[p].acyx < min) {
			min = temp[p].acyx;
                        i_min = p;
                }
                if (temp[p].acyx > max) {
			max = temp[p].acyx;
                        i_max = p;
                }
        }
        /* place min and max in its place and rearrange temp */
        ordered[0].acxy = temp[i_min].acxy;
        ordered[0].num_x = temp[i_min].num_x;
        ordered[0].acyx = temp[i_min].acyx;
        ordered[0].num_y = temp[i_min].num_y;
        ordered[1].acxy = temp[i_max].acxy;
        ordered[1].num_x = temp[i_max].num_x;
        ordered[1].acyx = temp[i_max].acyx;
        ordered[1].num_y = temp[i_max].num_y;
        size_ord = 2;
        for (p = 0; p < (num_shuffle-1); p++) {
                if (p >= i_min) {
                        temp[p].acxy = temp[p+1].acxy;
                        temp[p].num_x = temp[p+1].num_x;
                        temp[p].acyx = temp[p+1].acyx;
                        temp[p].num_y = temp[p+1].num_y;
                }
        }
        for (p = 0; p < (num_shuffle-2); p++) {
                if (p >= i_max) {
                        temp[p].acxy = temp[p+1].acxy;
                        temp[p].num_x = temp[p+1].num_x;
                        temp[p].acyx = temp[p+1].acyx;
                        temp[p].num_y = temp[p+1].num_y;
                }
        }
        /* place the rest of elements in their order */
        for (p = 0; p < (num_shuffle-2); p++) {
                q = 0;
                while (q <= (size_ord-2)) {
                        if ((ordered[q].acyx < temp[p].acyx) && (ordered[q+1].acyx >= temp[p].acyx)) {
                                /* insert element since we found its place */
                                size_ord = size_ord + 1; /* list increases by 1 element */
                                for (w = (size_ord-1); w > (q+1); w--) { /* rearrange ordered list */
                                        ordered[w].acxy = ordered[w-1].acxy;
                                        ordered[w].num_x = ordered[w-1].num_x;
                                        ordered[w].acyx = ordered[w-1].acyx;
                                        ordered[w].num_y = ordered[w-1].num_y;
                                }
                                /* insert element in its position */
                                ordered[q+1].acxy = temp[p].acxy;
                                ordered[q+1].num_x = temp[p].num_x;
                                ordered[q+1].acyx = temp[p].acyx;
                                ordered[q+1].num_y = temp[p].num_y;

                                break;
                        }
                        q = q + 1;
                }
        }
        /* place order in num_y */
        for (p = 0; p < num_shuffle; p++) {
                ordered[p].num_y = (num_shuffle - 1) - p;
        }

        /* get the rank of the original rank */
	rank = 0;
	result0_ptr->num_y = rank;  /* initialize at the bgining of the list */
	found = 0;
        for (p = 0; p < num_shuffle; p++) {
                if ((result0_ptr->acyx >= ordered[p].acyx) && (found == 0)) {
                        rank = (num_shuffle - 1) - p;
                        result0_ptr->num_y = rank;
                }
		else {
			found = 1;
		}
        }

        /* transfer ordered list to result_ptr */
        for (p = 0; p < num_shuffle; p++) {
                result_ptr[p].acxy = ordered[p].acxy;
                result_ptr[p].num_x = ordered[p].num_x;
                result_ptr[p].acyx = ordered[p].acyx;
                result_ptr[p].num_y = ordered[p].num_y;
        }

        /* calculate the probability and its error based on its rank */
	if (rank == 0) rank = 1;
        *prob_y_ptr = ((rank * 1.0)/((num_shuffle + 1) * 1.0));
	prob = *prob_y_ptr;
	*e_proby_ptr = ((prob * (1.0 - prob))/(num_shuffle * 1.0));

        /* We should free the memory occupied by the linked list */
        free(temp);
	free(ordered);

	return;

}

/****************************************************************
Print the copyright information.
*****************************************************************/
void print_cr(void)
{
 
        (void)fprintf(stdout, "\n \n \n *****************************************************************\n \n");
        (void)fprintf(stdout, "    (C) Copyright. All rights reserved. Juan M. Gonzalez. 2006.\n");
        (void)fprintf(stdout, "    Permission to use, reproduce, and distribute this program\n");
        (void)fprintf(stdout, "    or any of its parts should be granted by the author. \n \n");
        (void)fprintf(stdout, "    Juan M. Gonzalez, \n");
        (void)fprintf(stdout, "    IRNASE, P.O.Box 1052, 41080 Sevilla, Spain.\n");
        (void)fprintf(stdout, "    jmgrau@irnase.csic.es \n \n");
        (void)fprintf(stdout, " *****************************************************************\n \n \n");
 
        return;
 
}
 
/****************************************************************
Instructions for command-line arguments and program usage.
Prints the information and exits the program.
*****************************************************************/
void usage(char *program_name)
{
 
/* Help page for \'fingshuf1.c\'. */

(void)fprintf(stdout, "\n \n"); 
(void)fprintf(stdout, "%s statistically compares two communities based on fingerprinting results.\n \n", program_name);
(void)fprintf(stdout, "Usage is %s -[options]<string or number>\n", program_name);
(void)fprintf(stdout, "Input files includes a migration number and its density or proportion. Read the README file for format details.\n");
(void)fprintf(stdout, "Options are:\n");
(void)fprintf(stdout, "  -I<input file>   Input file from community 1 (X).\n");
(void)fprintf(stdout, "  -i<input file>   Input file from community 2 (Y).\n");
(void)fprintf(stdout, "  -O<output file>  Output file with the results.\n");
/* (void)fprintf(stdout, "  -A<integer>      Total abundance in the community to be analyzed.\n");
*/
(void)fprintf(stdout, "  -S<integer>      Number of shuffles to be performed. Default = 10000.\n");
(void)fprintf(stdout, "  -E<number>	  Error for considering bands as equal or different. Default = 0.05.\n");
(void)fprintf(stdout, "  -M<integer>      Abundance of the less relative abundant OTU. Default = 10.\n");
(void)fprintf(stdout, "  -t<integer>      Selects to get CPU time information. Default is No (0) and 1 will print it.\n");
(void)fprintf(stdout, "Example:\n");
(void)fprintf(stdout, "   %s -Iinput1 -iinput2 -Ooutput -S1000 -E4 -M20 -t \n \n", program_name);
exit(8);
}

